## See https://packaging.python.org/guides/single-sourcing-package-version/

'''A collection of functions for ipython use in the classroom

   -------   Time-stamp: <2021-07-02T17:50:29.307889-04:00 cws2>   ------  '''


__version__ = "0.0b2.dev32" # X.YaN.devM format for now




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Aug  1 10:32:35 2020
-----    Time-stamp: <2020-08-02T14:16:35.633361-04:00 hedfp>      -----

@author: (of mods) Carl Schmiedekamp
@Original Source: https://gist.github.com/thomasaarholt/b89b3f29ad82cd32176f2605834fa8ac


WARNING: Does not work in Jupyter Lab!

Inssstructions:
    1. pip install classroom_gizmos
    in notebook:
        from classroom_gizmos.jupyter_file_browser import box, get_file_path
        box    ## creates the file selection ipywidget box.
        ---
        path = get_file_path()  ## every time this cell is executed, the currently 
                                ## selected file's path is returned.

Instructions, using source:
    1. paste code in Jupyter notebook cell
    2. execute the cell and select file
    3. in another cell call get_file_path() which returns selected file.

"""

import ipywidgets as widgets
## Ref: https://gist.github.com/thomasaarholt/b89b3f29ad82cd32176f2605834fa8ac

import os
from pathlib import Path

newpath = None ## path to selected file or folder
newfile = None ## full path returned by get_file_path()

cwd = Path(os.getcwd())

FOLDERLABEL = '-------FOLDERS-------'
FILESLABEL = '-------FILES-------'

def get_folder_contents(folder):
    'Gets contents of folder, sorting by folder then files, hiding hidden things'
    folder = Path(folder)
    folders = [item.name for item in folder.iterdir() if item.is_dir() and not item.name.startswith('.')]
    files = [item.name for item in folder.iterdir() if item.is_file() and not item.name.startswith('.')]
    # print( 'get_folder_contents--')
    return [FOLDERLABEL] + sorted(folders) + [FILESLABEL] + sorted(files)

def go_to_address(address):
    address = Path(address)
    if address.is_dir():
        address_field.value = str(address)
        base.value = os.path.basename( str(address))
        select.unobserve(selecting, names='value')
        select.options = get_folder_contents(folder=address)
        select.observe(selecting, names='value')
        select.value = None
    # print( 'go_to_address--')

def newaddress(value):
    go_to_address(address_field.value)
    ##print( 'newaddress--')
        
def selecting(value):
    global newpath
    if value['new'] and value['new'] not in [FOLDERLABEL, FILESLABEL]:
        path = Path(address_field.value)
        newpath = path / value['new']
        if newpath.is_dir():
            go_to_address(newpath)
            
        elif newpath.is_file():
            #some other condition
            pass
        # print( 'selecting-- path:{} newpath: {}'.format( path, newpath))
        # print( '  types path: {}, newpath: {}, value[ new]: {}'.format( type( path), type( newpath), type( value['new'])))

def parent(value):
    new = Path(address_field.value).parent
    go_to_address(new)
    # print( 'parent--')
    
def load(value):
    global newfile
    ##filepath = path / value['new']
    # load using your favourite python software!
    newfile = newpath
    # print( 'load-- newpath: {}, newfile: {}'.format( newpath, newfile))

def get_file_path():
    '''User access to last file selected.'''
    # print( 'getFilepath-- newfile: {}, newpath: {}'.format( newfile, newpath))
    return newfile
    
    
address_field = widgets.Text(value=str(cwd))
up_button = widgets.Button(description='Up')
select = widgets.Select(options=get_folder_contents(cwd), rows=15, value=None)
load_button = widgets.Button(description='Select')
base = widgets.Text( str( os.path.basename( cwd)))
box = widgets.HBox([widgets.VBox([address_field, base, select]), widgets.VBox([up_button, load_button])])

select.observe(selecting, names='value')
up_button.on_click(parent)
#load_button.observe(load)
load_button.on_click(load)
address_field.on_submit(newaddress)

##in jupyter execute box
##box


# -*- coding: utf-8 -*-
"""
Handy Functions and variables/constants for use at iPython prompt
   call mine() for full list of functions and variables.

Created on Sat Feb  1 15:05:51 2020
     ------ Time-stamp: <2021-07-02T16:53:18.534724-04:00 hedfp> ------
     
@author: Carl Schmiedekamp

2020-04-14 /CS/ adding 'mine()' to list what has been setup by handies.py
                adding atan2P(y,x) to return positive angle.
2020-04-17 /CS/ adding astropy.units as 'u'
2020-07-13 /CS/ removing cdpy1d(), and cdWD()
2020-07-22 /CS/ making astropy and func_timeout optional.
2020-07-25 /CS/ added timeStampStr()
                changed from file date/time to timestamp file as mod. time
                for this module, also added the version from __init__.py.
                Added variable "__version__" which holds the version.
2020-07-29 /CS/ added isInstalled()
2020-08-13 /CS/ added condaEnvName() as a helper function.
2020-08-25 /CS/ added 'line magic' cls to fill screen with 23 blank lines.
                the 'Loading' message is not output if not in IPython or the
                interactive interpreter.
2020-09-02 /CS/ Now sets ipython number precision to 5 digits.
                  (Note: only for expression evaluation output.)
2020-09-13 /CS/ made the height parameter to figsize optional, with the default
                  being 0.75 % of width.
2020-09-22 /CS/ added VURound and round_sig functions,
                VURound needs work if uncertainty can have more than 1 sig.fig.
2020-10-16 /CS/ Add is_ipython(), is_interactive(), and is_commandline() to test
                running environment.
2020-10-30 /CS/ adding astropy.constants as c
2020-11-02 /CS/ added clsall() which clears terminal screen of all characters.
2021-05-05 /CS/ split round_sig into two function, round_sig_limit rounds to zero
                  values that are 'very small'. While round_sig( ) rounds to specified
                  sig.figs.
2021-05-10 /CS/ added is_defined(), which checks if name is defined.

#####

To Do List:
    - check all functions are in mine()
    - use consistant names ex: isIn vs is_in, get_osname vs osname
    - renamed functions from previous versions get aliases like "ii = isInstalled"
        for old names.
    - add function to timestamp current python file
    - Should the 'mine()' function be renamed to info()?

"""
  ## Hoping for some combatibility in python2.7
from __future__ import absolute_import, division
from __future__ import print_function, unicode_literals

  ## These includes should be available on most systems
import math
import codecs
from math import pi, sqrt, cos, sin, tan, floor
from math import acos, asin, atan, atan2, degrees, radians
from math import log, log10, exp

from random import randint

from platform import python_version
import sys, os.path, time, os


import glob, pathlib

from typing import List, Sequence, Any

  ## More for Python 2.7 compatibility
if sys.version_info[0] < 3:
    ## load some stuff so that the code can be more compatible with python3
    def input( str):
        '''' redefine input() for Python 2 so it is compatible with Python 3 version'''
        return raw_input( str)  ## raises warning in python3 but not used in python3



  ### Current rule of thumb: only handies imports from other modules.
from classroom_gizmos.BestByMinBefore import getCCode
from classroom_gizmos.import_install import importInstall as II
from classroom_gizmos.import_install import ckII


def clsall() -> None:
    '''Outputs the ASCII clear screen character.
    This usually deletes all the previoous text in the terminal.'''
    print( '\033[2J', end=None)

def is_online() -> bool:
    '''Returns True if connected to Internet.'''
    from urllib.request import urlopen
    
    try:
        urlopen( 'https://www.google.com/', timeout=20)
        return True
    except:
        return False

def is_ipython() -> bool:
    '''Return True if running in IPython.
    Ref: https://stackoverflow.com/questions/23883394/detect-if-python-script-is-run-from-an-ipython-shell-or-run-from-the-command-li
    '''
    try:
        __IPYTHON__
    except NameError:
        result = False
    else:
        result = True
    return result

# def is_interactive():
#     '''Return True if running interactive interpreter but not IPython.
#     Ref: https://stackoverflow.com/questions/23883394/detect-if-python-script-is-run-from-an-ipython-shell-or-run-from-the-command-li
#     '''
#     import inspect
    
#     if len( inspect.stack()) > 1 and not is_ipython():
#         result = True
#     else:
#         result = False
#     return result

def is_interactive() -> bool:
    '''Return true if interactive python, False if running from command line.
    Ref: https://stackoverflow.com/questions/2356399/tell-if-python-is-in-interactive-mode
    '''
    import sys
    try:
        if sys.ps1:
            inter = True
    except AttributeError:
        inter = False
        if sys.flags.interactive: inter = True
    return inter


def is_commandline() -> bool:
    '''Returns True if running from command line (not ipython, and not interactive
    interpreter.).
    Ref: https://stackoverflow.com/questions/23883394/detect-if-python-script-is-run-from-an-ipython-shell-or-run-from-the-command-li
    '''
    import inspect
    
    if len( inspect.stack()) == 1 :
        result = True
    else:
        result = False
    return result

def is_collab() -> bool:
    '''Returns True if running in Google Colab.
    Ref: https://stackoverflow.com/questions/53581278/test-if-notebook-is-running-on-google-colab'''
    import sys
    return 'google.colab' in sys.modules

def round_sig_limit( x: float, sig: int=6, small_value: float=1.0e-9) -> float:
    ''' Round numbers to sig figs, and zero if less than min. value.
      Modified to to work without limit and created new round_sig_limit()
     Ref: https://stackoverflow.com/questions/3410976/how-to-round-a-number-to-significant-figures-in-python
     '''
    if abs( x)<abs(small_value):  ## return zero for smaller values
        return 0.0
    else:
        return round(x, sig - int(floor(log10(abs(x)))) - 1)

def files_globber( wild_card_file_list: List[str], full: bool=True) -> List[ str]:
    '''
    
      expand wild cards in the files in wild_card_file_list
       and create a single list of path names.
       
       if full flag is True, then full paths are returned,
         otherwise the base names (without directory parts) 
         of the files are returned.
       
       Any duplicate names are removed before returning the list
         of files found.
       For example:
      files_globber( [ './*.txt', '*.txt', ../dir/*.txt'])
       could specify the same list of files, 3 different ways, 
       the duplicates are removed and a file name appears
       once in the output list.  However, different paths to the
       same file, such as those created with the ln command are
       all returned.
    '''

    
    
    full_path_list = []      ## list of full path names found
    filesRel = []   ## wild card expanded file list
    base_name_list = [] ## list of base file name (file part) corresponding
                        ##   to paths in files
                        
    ## Catch single string instead of a list of strings
    # print( f'DBug type of arg is { type(wild_card_file_list)}')
    if type( wild_card_file_list) == str:
        # print( 'DBug converting string to list.')
        file_list = [ wild_card_file_list ]
    else:
        file_list = wild_card_file_list
    
    for f in file_list:
        # print( f'DBug: current file is {f}')
        filesRel += glob.glob( f)
    # print( '\nDBug filesRel: {}'.format( filesRel))
    
    for f in filesRel:
        path = pathlib.Path(f)
        # print( '\nDBug path: {}, Abs: {}'.format(
        #     path, path.absolute()))
        fabs = path.absolute()
        path = str(fabs)
        full_path_list.append( path)
        base =  os.path.basename( path)
        base_name_list.append( base)
    print( '\nDBug files: {}'.format( full_path_list))
    print( f'\nDBug base names: { base_name_list}')
    if full:
        return list( set( full_path_list))  ## remove duplicates and return list
    else:
        return list( set( base_name_list))  ## remove duplicates and return list


def round_sig(x: float, sig: int=6):
    ''' Round numbers to sig figs.
      Modified to to work without limit and created new round_sig_limit()
     Ref: https://stackoverflow.com/questions/3410976/how-to-round-a-number-to-significant-figures-in-python
     '''
    return round(x, sig - int(floor(log10(abs(x)))) - 1)



def VURound( value: float, uncertainty: float, undig: int=1) -> str:
    '''
    Returns a string with rounded value and uncertainty.
    Round value based on uncertainty.
    Uncertainty is rounded up to "undig" sig.figs.
      'undig' is currently fixed at 1
      Based on Javascript function by Carl Schmiedekamp
    '''
    from math import copysign, log10, floor, ceil
    
    #soecial case: uncertanity is zero: return value as string.
    if uncertainty==0:
        return str( value)
    
    sign = copysign( 1,value)
    
    valin = abs( value)
    uncin = abs( uncertainty)
    
    dec = log10(uncin)
    # print('DBug uncin, dec, uncertainty', uncin, dec, uncertainty)
    dec = floor(dec+1)-1
    
    uncmant=uncin*pow(10.0,-1*dec)
    
    if uncmant > 9: ##Fix if rounding up adds another digit.
        dec = dec + 1
        uncmant=uncin*pow(10.0,-1*dec);
    
    if round( uncmant) != uncmant:  ##round larger if sigmant is not integral
        uncmant=int( uncmant+1);
    
    ## for values between .1 and 9999. use 2 part output (also equal to zero)
    if (valin >= .1 and valin <=9999.) or (valin == 0.0):
        if dec > 0:
            nd = 0
        else:
            nd = -dec
            
        fmtstr = '{'+':22.{}f'.format( nd)+'}'
#        print( 'DBug: fmtstr: {}'.format( fmtstr))
        
        if sign < 0:
#            val3p=str(int(valin*pow(10,-dec)+0.5)*pow(10,dec)*sign)
            val3p=fmtstr.format( int(valin*pow(10,-dec)+0.5)*pow(10,dec)*sign).strip()
        else:
#            val3p=str(int(valin*pow(10,-dec)+0.5)*pow(10,dec))
            val3p=fmtstr.format( int(valin*pow(10,-dec)+0.5)*pow(10,dec)).strip()
            
#        print( 'DBug: valin: {}, dec: {}, val3p: {}, nd: {}'.format( valin, dec, val3p, nd))
        
        unc3p = fmtstr.format( uncmant*pow(10,dec) ).strip()
        
        return ('{} '+u"\u00B1"+ ' {}').format( val3p, unc3p)
        
    else:  ##  otherwise use 4 part output
        
        val=int(valin*pow(10,-dec)+0.5)*pow(10,dec)*sign
        if valin >= uncin:
            exp=ceil(log10(valin))-1
        else:
            exp=ceil(log10(uncin))-1
        
        
        nd = -( dec - exp)
        
        if nd < 0:
            nd = 0
            
        fmtstr = '{'+':22.{}f'.format( nd)+'}'
#        print( 'DBug: fmtstr: {}'.format( fmtstr))

        val = fmtstr.format( val*pow(10,-exp)).strip()

        unc=str(uncmant*pow(10,dec-exp));
        unc = fmtstr.format( uncmant*pow(10,dec-exp)).strip()

#        print( 'DBug: val: {}, unc: {}, exp: {}, nd: {}'.format( val, unc, exp, nd))
        
        return ('({} '+u"\u00B1"+ ' {})' + u"\u00D7" +'10^' + '{}').format( val, unc, exp)



def randomLetter() -> str:
    '''
    Generate a single random uppercase ASCII letter

    Returns
    -------
    TYPE
        char.
    '''
    import random
    import string    
    return random.choice(string.ascii_uppercase)

def randomElement( List: Sequence[ Any]=(-1,1)) -> Any:
    '''
    return random element of list/sequence

    '''
    index = randint(0, len( List)-1)
    return List[ index]


def timeStampStr() -> str:
    '''Returns a string with the current time in ISO format.'''
    import datetime
    import time
    
    dt = datetime.datetime.now()
    epoch = dt.timestamp() # Get POSIX timestamp of the specified datetime.
    st_time = time.localtime(epoch) #  Get struct_time for the timestamp. This will be created using the system's locale and it's time zone information.
    tz = datetime.timezone(datetime.timedelta(seconds = st_time.tm_gmtoff)) # Create a timezone object with the computed offset in the struct_time.
    return dt.astimezone(tz).isoformat()

## math.comb is in Python 3.8 
if hasattr( math, "comb" ):
    from math import comb
    from math import comb as nCr
else:
    def nCr(n: int, r: int) -> int:
        ''' calculate number of combinations of r from n items.'''
        f = math.factorial
        return f(n) // f(r) // f(n-r)
    comb = nCr   

def relative_to_full_path( rel_path: str) -> str:
    '''Converts file path, relative to this module's directory, to a full path.
    Ref: https://stackoverflow.com/questions/3718657/how-do-you-properly-determine-the-current-script-directory
    '''
    import inspect
    filename = inspect.getframeinfo(inspect.currentframe()).filename
    path = os.path.dirname(os.path.abspath(filename))
    path = os.path.join( path, rel_path)
    return path

def getTS( rel_path: str) -> str:
    '''Reads first line of specified file, which is expected to be a time-date
    string.
    This is mostly for internal package use.'''
    # here = os.path.abspath(os.path.dirname(__file__))
    # with codecs.open(os.path.join(here, rel_path), 'r') as tsf:
    fpath = relative_to_full_path( rel_path)
    # print( f'DBug: rel_path: { rel_path}, fpath: {fpath}')
    with codecs.open( fpath, 'r') as tsf:
        lines = tsf.read()
    linelist = lines.splitlines()
    return linelist[0]


def read( rel_path: str) -> str:
    '''Reads text file.'''
    # here = os.path.abspath(os.path.dirname(__file__))
    # with codecs.open(os.path.join(here, rel_path), 'r') as fp:
    fpath = relative_to_full_path( rel_path)
    # print( f'DBug: fpath: {fpath}')
    with codecs.open( fpath, 'r') as fp:
        return fp.read()

def get_version(rel_path: str) -> str:
    '''Reads version from specified file. The file path is relative
    to this file. It looks for line that starts with "__version__"'''
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

def is_defined( name:str) -> bool:
    '''
    Tests if string in name is a defined variable, function, etc.

    Parameters
    ----------
    name : str
        Name of object.

    Returns
    -------
    bool
        True if contents of name is name of defined object, else False.

    '''

    return name in locals()



## define cls magic and set precision if in IPython.
if is_ipython() and is_interactive():
    ### set default float precision
    ### Ref: https://stackoverflow.com/questions/10361206/how-to-run-an-ipython-magic-from-a-script-or-timing-a-python-script
    from IPython import get_ipython
    from IPython.core.magic import register_line_magic
            
    ipython = get_ipython()
    ipython.magic( 'precision %0.5g')
            
    ### define clear screen magic (%cls generates 23 blank lines)
    @register_line_magic 
    def cls(line) -> None: 
        '''Defines a 'clear screen' line magic'''
        print( 23*'\n') 
        return

    del cls ## must delete function to make magic visible.
    
    

######  'Welcome Message' on loading  ######

def condaEnvName() -> str:
    '''Gets environment name from python's path.
    Ref:
        https://stackoverflow.com/questions/36539623/how-do-i-find-the-name-of-the-conda-environment-in-which-my-code-is-running'''

    # return Path(sys.executable).as_posix().split('/')[-3]
    return sys.exec_prefix.split(os.sep)[-1]

def ipythonversion() -> str:
    '''gets IPython version or returns None.'''
    try:
        import IPython
        ipv = IPython.version_info
        version = ''
        for n in ipv:
            version += str( n) + '_'
        return version[:-1]  ## remove last '_'
    except:
        return None
 
def in_ipynb() -> bool:
    '''True if running in Jupyter notebook.
    Ref: https://exceptionshub.com/how-can-i-check-if-code-is-executed-in-the-ipython-notebook.html
    Ref: https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook.
    Currently not definitive in result; returns True in Spyder IPython console.'''

    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter


def ck_installed( pkgname: sys.modules) :
    ''' Checks if package is installed,
        If installed the package is returned,
        Else tries to install pkgname and returns package if installed.
        pkgname is a string with name of package as used in import stmt.
        If install fails, returns None.
        Should be a duplicate of import_install.is_installed().
        Typical Usage:
            astropy = isInstalled( 'astropy') 
            if astropy == None:
                ... 
            else:
                from astropy import units as u
        '''
    try:
        pkg = __import__( pkgname)
        return pkg
    except Exception:
    	return None
ii = ck_installed ## alias

def is_installed( pkgname: sys.modules) -> bool:
    '''Returns True if package is installed.'''
    try:
        __import__( pkgname)
        return True
    except Exception:
    	return False

## check if packages for sound are installed and load once
IPython = ii( 'IPython')
beepy = ii( 'beepy')
 # beepy = None  ## testing
playsoundPkg = ii( 'playsound')
 # playsoundPkg = None; ## testing
simpleaudio = ii( 'simpleaudio')
 # simpleaudio = None ## testing

def beepsound() -> None:
    '''Plays a 'beep' sound when called if a sound package is found, or
    tries terminal beep.
    Sound packages that can be used:
        beepy
        simpleaudio
        Ipython.display.Audio'''
    ## use beep sounds from various possible packages
    if beepy != None:
        beepy.beep( 1)
    elif playsoundPkg != None and os.path.isfile( 'A-Tone-His_Self-1266414414.wav'):
        playsoundPkg.playsound( 'A-Tone-His_Self-1266414414.mp3')
    elif simpleaudio != None and os.path.isfile( 'mixkit-relaxing-bell-chime-3109.wav'):
        wave_obj = simpleaudio.WaveObject.from_wave_file( 'A-Tone-His_Self-1266414414..wav')
        play_obj = wave_obj.play()
        play_obj.wait_done()
    elif IPython != None:
        try:
            # IPython.display.Audio( 'http://www.soundjay.com/button/beep-07.wav', autoplay=True)
            IPython.display.Audio( 'audio/tone_440.wav', autoplay=True)
            # import numpy
            # sr = 22050 # sample rate
            # T = 2.0    # seconds
            # t = numpy.linspace(0, T, int(T*sr), endpoint=False) # time variable
            # x = 0.5*numpy.sin(2*numpy.pi*440*t)  
            # IPython.display.Audio(x, rate=sr) # load a NumPy array
        except ValueError:
            print( '\a')
    else:
        print( '\a')   ## probably won't work, but last attempt.


    
## Get timestamp for package, updated by setup.py.
timestamp = getTS( 'timestamp.txt')
date  = timestamp[0:10]
# print('\nDBug: date: {} TS:\n{}'.format( date, timestamp))

__version__=get_version("__init__.py")

def call( cmd: str):
    import subprocess
    '''Modeled after call function in NANOGrav Sprinng 2020 workshop.
    call() just executes the command in the shell and displays output.
    
    Could have security issues. See subprocess documentation.'''
    subprocess.call( cmd, shell=True)


def mine() -> None:

    '''List (prints) the functions and variables defined in handies.py'''
    
    print('\n classroom_gizmos.handies ver: {}, modified {}'.format(
        __version__, date))
    if is_installed( 'astropy'):
        print('Defining:\n     nowMJD(); mjd2date(), date2mjd(),')
        print('     astropy.units as "u", i.e. u.cm or u.GHz')
        print('     astropy.constants as "c", i.e. c.c or c.au')
        print('     angular_separation() ➞ angular separation between two RA,DEC coords.')
    else:
        print( '** astropy not available; MJD functions, u (units) and c (constants) are not available.')
    print('     cdbn(), osname(), hostname(), call(),')
    if is_installed( 'PyQt5'):
        if is_installed( 'func_timeout'):
            print('     select_file(), select_file_timeout( timeout={}),'.format( sfTimeOut))
        else:
            print('** func_timeout not available; select_file_timeout() ignores timeout.')
            print('     select_file()')
        
    else:
        print( '** PyQt5 is not available; select_file() and select_file_timeout() are not defined.')
    print('     rad(), deg(), sinD(), cosD(), tanD(), asinD(),\n     acosD(), atanD(), atan2D(), atan2P()')
    print("       'D' and 'P' functions work with degrees.")
    print("       'P' inverse functions return only positive angles.")
    print("     Defines nCr() and comb() or imports from math if available.")
    print('     greeks  ➞ a string with greek alphabet.')
    print('     cls or %cls; ipython \'magic\' writes 23 blank lines to clear an area of the screen')
    print('     clsall() function which removes previous text on screen by outputing ascii code.')
    print('     pltsize( width) ➞ resizes plots in matplotlib, units are inches')
    print('     timeStampStr() ➞ returns a readable timestamp string.')
    print('     ck_installed( pkgNameStr) ➞ returns package or None if not installed.')
    print('     is_installed( pkgNameStr) ➞ returns True if package is installed.')
    print('     is_ipython() ➞ True if running in ipython.')
    print('     is_interactive() ➞ True if in interactive interpreter but not in ipython.')
    print('     is_commandline() ➞ True if running from commandline;not interactive nor ipython.')
    print('     is_colab() ➞ True if running from Google Colab.')
    print('     VURound( value, uncertainty) ➞ Rounds value based on uncertainty.')
    print('     round_sig( value, sigfigs) ➞ Rounds value to specified sig.figs.')
    print('     round_sig_limit( value, sigfigs, limit) ➞ Rounds value to specified sig.figs or zero.')
    print('     count_down() ➞ Counts down the specified number of seconds.')
    print('     beepsound() ➞  trys to make sound.')

    print('     randomLetter() ➞ a random uppercase ASCII letter.')
    print('     randomElement( List) ➞ returns random element from list.')
    
    print('     external_addresses()  ➞  external names and IPs')
    print('     externalIP() ➞ external IP')
    print('     local_addresses() ➞ local name and IP')
    
    print('  Functions for portable code:')
    print('     osname(), username(), computername(), pythonversion(),')
    print('     condaversion(), ipythonversion(), in_ipynb(), is_installed()')
    print('     ck_installed(), or ii(),  II()' )
    
    print()

    print('From random imports randint( min, max)')
    print('From BestByMinBefore imports getCCode()')
    print('From import_install imports import_install as II and imports ckII')
    
    print('From math imports:\n     pi, sqrt, degrees, radians,\n     cos, sin, tan, atan2, asin, acos, atan, and\n' + 
      '     log, log10, exp')

    print( '\n     mine() ➞ lists what handies.py defines.')

    print( '\nRequires astropy, PyQt5, and func_timeout packages for full functionality.')
    print( 'beepsound() can use these audio packages if available: beepy, playsound, simpleaudio')
    
##Ref: https://ipython.readthedocs.io/en/stable/interactive/magics.html#magic-precision
    # prhints = ('Hint:\n' + 
    #       '     "%precision %g" or %.6g for better number formats.')
    # print(prhints)


def cosD( ang: float) -> float:
    '''Return cosine of angle in degrees.'''
    import math as m
    return m.cos( m.radians( ang))

def sinD( ang: float) -> float:
    '''Return sine of angle in degrees.'''
    import math as m
    return m.sin( m.radians( ang))

def tanD( ang: float) -> float:
    '''Return tangent of angle in degrees.'''
    import math as m
    return m.tan( m.radians( ang))

def asinD( val: float) -> float:
    '''Return inverse sine, in degrees, of val.'''
    import math as m
    return  m.degrees( m.asin( val))

def acosD( val: float) -> float:
    '''Return inverse consine, in degrees, of val.'''
    import math as m
    return m.degrees( m.acos( val))


def atanD( val: float) -> float:
    '''Return inverse tangent, in degrees, of val.'''
    import math as m
    return m.degrees( m.atan( val))


def atan2D( y: float, x: float) -> float:
    '''Return inverse tangent, in degrees, of y/x. (-180 ... 180)'''
    import math as m
    return m.degrees( m.atan2( y, x))

def atan2P( y: float, x: float) -> float:
    '''Return inverse tangent, in degrees, of y/x. (0 .. 360)'''
    ang = atan2D( y, x)
    if ang <0:
        ang = ang + 360
    return ang

rad = radians  ## alias the conversion functions
deg = degrees

greeks = ' Α Β Γ Δ Ε Ζ Η Θ Ι Κ Λ Μ Ν Ξ Ο Π Ρ Σ Τ Υ Φ Χ Ψ Ω  α β γ δ ε ζ η θ ι κ λ μ ν ξ ο π ρ σ τ υ φ χ ψ ω '

def cdbn( dirname: str, sub=None) -> str:
    '''cd to directory (or subdirectory) who's path is in Env.Var. who's name is passed as
       first argument.  2nd argument specifies a subdirectory relative to
       that named directory. On successful change, returns path of new CWD.'''
    import os
    """ Does cd to directory named in specified environment
        variable.
        Returns current directory (as string) if no error.
        If Error, outputs error msg. and returns False.
        --
        If sub is specified, tries to cd to that subdirectory
        after the cd to the contents of env. var.
    """
    try:
        dir = os.environ[ dirname]
    except KeyError:
        print( dirname, ': not an env. var.')
        return False
    else:
        if os.access( dir, os.R_OK):
            os.chdir( dir)
            if sub==None:
                return os.getcwd()
            else:
                if os.access( sub, os.R_OK):
                    os.chdir( sub)
                    return os.getcwd()
                else:
                    print( '{}: No access!'.format( 
                            os.path.join(dir,sub), dir))
            return False

        else:
            print('[{}] --> {} : No access!'.format( dirname, dir))
            return False



try:   ## These are only defined if astropy is installed.
    import astropy
    from astropy import units as u
    from astropy import constants as c
    import astropy.coordinates.angles as aca
    
    def nowMJD() -> float:
        '''Convert current time to MJD'''
        from astropy.time import Time
        return Time.now().to_value( 'mjd')
    
    def mjd2date( mjd: float) -> str:
        '''Convert MJD to a civil utc date/time'''
        
        from astropy.time import Time
        time = Time( mjd, format='mjd')
        return time.to_datetime()
    
    def date2mjd_( civildate: str) -> float:
        '''Convert specified time to MJD.
        The string in civildate must be recognized by astropy.time.Time,
        and is assumed to be UCT time.
    
        Value returned is float
        
        Usage:
            date2mjd( '2020-05-16T14:10') returns 58985.59027777777778'''
        
        from astropy.time import Time
        return Time( civildate).to_value( 'mjd', 'long')
        
    def date2mjd( civildate: str) -> float:
        '''Convert specified time to MJD.
        The string in civildate must be recognized by astropy.time.Time,
        and is assumed to be UCT time.
        As a special case, if the date/time is followed by UTC offset
        then that is used to shift the date/time to UTC.
        Example: 2019-09-30 20:54:54-04:00 corresponds to 
        UTC of 2019-10-01 00:54:54
    
        Value returned is float
        
        Usage:
            date2mjd( '2020-05-16T14:10') returns 58985.59027777777778
            or
            date2mjd( '2020-05-16T14:10-04:00') returns 58985.756944444444446
            '''
        
        cd = civildate
        offset = 0
        ## check for special case:
        if cd[-3]==':' and ( cd[-6]=='-' or cd[-6]=='+') :
            offset = (int( cd[-5:-3]) + int( cd[-2:])/60 )/24
            if cd[-6]=='+':
                offset = -offset
            cd = cd[:-6]
        
        from astropy.time import Time
        return Time( cd).to_value( 'mjd', 'long')+offset
    
    def angular_separation( RA1: str, DEC1: str, RA2: str, DEC2: str
                           ) -> aca.Angle:
        '''
        Returns angle separation between two coordinates in Astropy 'Angle' units.
    
        Parameters
        ----------
        RA1 : string
            RA coordinate (ex: '5h23m34.5s')
        DEC1 : string
            DEC coordinate (ex: '-69d45m22s')
        RA2 : string
            RA coordinate (ex: '0h52m44.8s')
        DEC2 : string
           DEC coordinate (ex: '-72d49m43s')
    
        Returns
        -------
        Separation angle in Astropy Angle units.
    
        '''
        
        from astropy.coordinates import SkyCoord
        # c1 = SkyCoord('5h23m34.5s', '-69d45m22s', frame='icrs')
        # c2 = SkyCoord('0h52m44.8s', '-72d49m43s', frame='fk5')
        c1 = SkyCoord( RA1, DEC1)
        c2 = SkyCoord( RA2, DEC2)
        sep = c1.separation(c2)
        
        return sep
    
except ImportError:
    print( 'mjd and anglular_separation functions not defined because astropy is not available.')


def pltsize( w: float, h: float=None, dpi: int=150) -> None:
    '''set plot size (matplotlib), size in notebook depends on resolution and
    browser settings. However, doubling the values should double the size.
    dpi is dots-per-inch which also changes plot size in notebooks.
    Default height is .75 times the width.'''
    import matplotlib
    if h==None:
        h = 0.75*w
    matplotlib.rc('figure', figsize=[w,h], dpi=dpi)

def osname() -> str:
    '''Returns name of operating system that Python is running on.'''
    try:
        os = __import__( 'os')
        return os.name
    except ImportError:
        return None


def username() -> str:
    '''Get current username or return "None".'''
    # import getpass
    try:
        getpass = __import__( 'getpass')
        return( getpass.getuser())
    except ImportError:
        return None

def computername() -> str:
    '''Get hostname of current computer or return "None".'''
    try:
        socket = __import__( 'socket')
        return socket.getfqdn()
    except ImportError:
        return None
hostname = computername

def external_addresses() -> List[ str]:
    '''Returns a 3 element list of info about external IP/Hostname.
      ( external hostname,  list of aliased hostnames, list of aliased IPs)
      
    '''
    from requests import get
    import socket
    
    verbose = False  ## set to True for debugging output
    
    ip = 'IP address not found.' ##  To mark failed attempt to get ip address
    
    try:
        ip = get( 'https://checkip.amazonaws.com').text
        if verbose: 
            print( f'from amazonaws.com ip is {ip}')
    except Exception:
        try:
            ip = get( 'https://api.ipify.org').text
            if verbose: 
                print( f'from ipify.org ip is {ip}')
        except Exception:
                ip = get( 'https://ident.me').text
                if verbose: 
                    print( f'from ident.me ip is {ip}')

    else:
        if verbose: 
            print('\nMy public IP address is: {}\n'.format(ip))
    
    ip = ip.rstrip()   ## remove any whitespace at ends of string.

    
    ##  Ref: https://www.programcreek.com/python/example/2611/socket.gethostbyaddr
    if verbose: 
        print( f'DBug ip is {ip}, and its type is {type(ip)}')
    
    try:
        results = socket.gethostbyaddr( ip)
    except Exception as err:
        results = ( str( err), [ None], [ ip])
        
    return results

def local_addresses() -> List[str]:
    ''' returns tuple of local hostname and local IP address'''
    import socket
    hostname = socket.gethostname()
    IP = socket.gethostbyname(hostname)
    return ( hostname, IP)

def externalIP() -> str:
    '''Returns external IP address via call to external_addresses().'''
    return external_addresses()[2][0]

def pythonversion() -> str:
    '''Get version of python or return "None".'''
    try:
        platform = __import__( 'platform')
        return platform.python_version()
    except ImportError:
        return None

def condaversion():
    '''Get version of conda, or return "None".
    Usually returns None!'''
    try:
        conda = __import__( 'conda')
        return conda.__version__
    except ImportError:
        return None  ## to be consistent with the others.

def count_down( t: int=10) -> None:
    '''Count down by seconds. 't' is the number of seconds.
    If beepy is installed, then a sound is made when countdown ends.
    Refs:  
    https://pypi.org/project/beepy/#description
    https://www.codespeedy.com/how-to-create-a-countdown-in-python/
    https://stackoverflow.com/questions/25189554/countdown-clock-0105'''
    while t > 0:
        mins, secs = divmod( int( t), 60)
        timeformat = '{:02d}:{:02d} '.format(mins, secs)
        print(timeformat, end='\r')
        time.sleep(1)
        t -= 1
    print('\a\n\nDone!\n\n')
    beepsound()



#>>> File hash code functions 
def base10toN( num: int, n: int) -> str:
    """Change 'num'  to a base-n number.
    Up to base-38 is supported without special notation.
    should be file name safe.
    
     0 <= num  (i.i. non-negative)
     2 <= n <= 38  implemented number bases
    invalid values result in '???' being returned.
    
    C.S.:extended from base 36 to 38
    
    Ref: https://code.activestate.com/recipes/65212-convert-from-decimal-to-any-base-number/
    """
    if num < 0:
        return '???'
    if n < 2 or n > 38:
        return '???'
        
    num_rep={10:'a', 11:'b', 12:'c', 13:'d', 14:'e', 15:'f',
                     16:'g', 17:'h', 18:'i', 19:'j', 20:'k',
                     21:'l', 22:'m', 23:'n', 24:'o', 25:'p',
                     26:'q', 27:'r', 28:'s', 29:'t', 30:'u',
                     31:'v', 32:'w', 33:'x', 34:'y', 35:'z',
                     36:'-', 37:'_'}
    alphalen = 38  ##  number of characters in alphabet
    new_num_string=''
    current=num
    while current!=0:
        remainder=current%n
        if alphalen > remainder > 9:  ## 10 to alphalen - 1 case
            remainder_string=num_rep[remainder]
        elif remainder>=alphalen:     ## error condition
            remainder_string='('+str(remainder)+')'
        else:                         ## 0 to 9 case
            remainder_string=str(remainder)
        new_num_string=remainder_string+new_num_string
        current=int(current/n)
    return new_num_string

def hashstrlist( strlist: List[ str], nc: int=5, ordered: bool=False) -> str:
    '''
    Creates file systems safe hash code of length nc from the list of
    strings in strlist. Typically, strlist is a list of file names.
    The order of the strings does not affect the hash generated unless ordered is True.
    
    uses base10toN
    2019-02-16 /CS/ initial function version
    2021-05-09 /CS/ added 'ordered' flag.
    '''
    import hashlib
    tmplist = strlist.copy()
    if not ordered:
        tmplist.sort()
#    print( 'DBug:\n', strlist,'\n-----\n', tmplist)
    tmp = "".join( tmplist)
#    print( 'DBug:\n', tmp)
    resulthash = int(hashlib.sha256(tmp.encode('utf-8')).hexdigest(), 16) 
    # print( 'DBug resulthash:', resulthash)
    result = resulthash % 38**(nc)
    # print( 'DBug: result:', result)
    result = base10toN( result, 38).upper()
    # print( 'DBug: result:', result)
    
    ## zero fill if length is less than nc
    while len( result) < nc:
        result = '0' + result
    return result

#>>> END File hash code functions        


try:
    import PyQt5  ## used in fdtest module, skip these definitions if Qt5 not available.
    
    ## default timeout (in s) for select_file():
    sfTimeOut = 95

    def select_file( ) -> str:
        ''' Uses fdtest.py to browse for file and return its path.
        WARNING: This function will 'hang' until a file is selected.'''
        
        from classroom_gizmos.fdtest import gui_fname        
        return gui_fname()

    def select_file_timeout( timeout: float=sfTimeOut):
        ''' Uses fdtest.py to browse for file and return its path.
        If the fdtest call takes longer than 'timeout' seconds,
        the attempt is cancelled and None is returned.
        The argument, timeout, sets the number of seconds before time out.
        Note: if func_timeout package is not available, the no timeout
        is used (select_file() is called.)
        '''
        ## trying to do auto-install of func_timeout
        try:
            from func_timeout import func_timeout, FunctionTimedOut
        except ImportError:
            from classroom_gizmos.import_install import ii
            func_timeout = ii( 'func_timeout')
            try:
                from func_timeout import func_timeout, FunctionTimedOut
            except Exception:
                print( 'Timeout is not available.')
                return select_file()
            
        from classroom_gizmos.fdtest import gui_fname
        
        try:
            filename = func_timeout( timeout, gui_fname, args=()).decode("utf-8")
        except FunctionTimedOut:
            outstr = 'select_file cound not complete within '
            outstr += '{} seconds.\n'.format( timeout)
            print( outstr)
            #raise e
            filename = None
            
        return filename
except ImportError:
    print( 'select_file functions not defined because PyQt5 is not available.')

user = username()  ## only output version info for me.
if ( user == 'cws2' or user == 'hedfp') and +\
    is_interactive() and not in_ipynb():
    print( "Loading Carl's handies.py ver: {} {}; Python:{};\n   environment: {}; IPython: {}; is interactive: {};\n is ipython: {}".format( 
            __version__, date, python_version(), condaEnvName(), ipythonversion(), is_interactive(), is_ipython() ) )




if __name__ == "__main__":
    mine()
    
    
    



from sys import executable, argv
from subprocess import check_output
from PyQt5.QtWidgets import QFileDialog, QApplication

def gui_fname(directory='./'):
    """Open a file dialog, starting in the given directory, and return
    the chosen filename"""
    # run this exact file in a separate process, and grab the result
    file = check_output([executable, __file__, directory])
    return file.strip()

if __name__ == "__main__":
    directory = argv[1]
    app = QApplication([directory])
    fname = QFileDialog.getOpenFileName(None, "Select a file...", 
            directory, filter="All files (*)")
    print(fname[0])


''' calc minutes before for BestBuy, In-class, etc. questions
returns a string that is the perl code fragment for specific cases.xs
2020-09-25 /CS/ revised wording in getCCode() prompts.'''
## ----- Time-stamp: <2021-07-02T17:59:33.885856-04:00 hedfp> -----

 
from __future__ import absolute_import, division, print_function, unicode_literals
 
import time
from math import sqrt
 
##############################################

### Default last days for classes:
duedates = { 'PHYS250.001':'2021-05-04 23:59:00', 'PHYS251': '2021-05-04 23:59:00',
             'PHYS250.002':'2021-05-04 23:59:00'}

### Added hours to end of full points time for safety margin.
nudge=1.9   ## h added to be 'safe'

### Usual times for credit to decrease
InC_hours = 12
BBP_hours = 24*7

### Usual minimum credit as fraction of full credit
minCreditF = 0.2

### Some time conversions
hr = 3600  ## 1 hour is 3660 s
day = 86400 ## 1 day is 87840 s
week = 604800 ## 1 week is 604880 s

def BB():
    print( '\nFunctions in BestByMinBefore:\n')
    print( 'decCredit( decStart, hoursToMinimum=BBP_hours, fullCreditSubs=5, minCreditSubs=10,')
    print( '          minCreditFraction=minCreditF, assignmentDue=duedates[ "PHYS250.001"]),')
    print( '          nudge=1.9')
    print()
    print( 'BBPdecCredit( decStart, hoursToMinimum=7*24)')
    print()
    print( 'InCdecCredit( decStart, hoursToMinimum=24)')
    print()
    print( 'HWdecCredit( )')
    print()
    print( 'getCCode() --> prompts user for parameter values.')
    print()
    print( '\n Helpers:')
    print( 'localTimeToGMTepoch( localTime)')
    print()
    print( 'minBeforeToLocal( minBefore, assignmentDue=duedates[ "PHYS250.001"])')
    
    

def localTimeToGMTepoch( localTime):
    '''
    

    Parameters
    ----------
    localTime : a date string in format "%Y-%m-%d %H:%M" or "%Y-%m-%d %H:%M:%S"
            ( yyyy-mm-dd hh:mm  or yyyy-mm-dd hh:mm:ss  )
        for a local (Eastern time) time

    Returns
    -------
    GMT epoch of that time
    '''
    
    try:
        tuple = time.strptime( localTime, "%Y-%m-%d %H:%M")
    except ValueError:
        try:
            tuple = time.strptime( localTime, "%Y-%m-%d %H:%M:%S")
        except ValueError:
            raise
            
    # print( 'DBug: tuple: {}'.format( tuple))
    
    epochFromLocal = time.mktime( tuple)  ## seconds since 1970-01-01 00:00:00 GMT
    # print( 'DBug: epoch for {} is {}'.format(
    #     localTime, epochFromLocal))

    return epochFromLocal

def minBeforeToLocal( minBefore, assignmentDue=duedates[ 'PHYS250.001']):
    '''
    Convert minutes before assignment is due to a date.

    Parameters
    ----------
    minBefore : integer
        minutes before the assignment is due.
    assignmentDue : string
        date/time string when assignment is due.
    Returns
    -------
    date/time string for minBefore minutes before assignmentDue

    '''
    beforeDueEpoch = localTimeToGMTepoch( assignmentDue) - minBefore*60  ## in seconds
    
    return time.asctime( time.localtime( beforeDueEpoch))


def decCredit( decStart, hoursToMinimum=BBP_hours, fullCreditSubs=5, minCreditSubs=10,
              minCreditFraction=minCreditF, assignmentDue=duedates[ 'PHYS250.001'], nudge=nudge):
    '''
    General decreasing credit perl string generator.

    Parameters
    ----------
    decStart : (a date string in format "%Y-%m-%d %H:%M:%S")
        This is the date/time when the credit for an correct answer starts to decrease.
        if the string is empty or parameter == None, then credit does not decrease
          with time.
        A date string, example:
            2020-08-31 23:59:00
    hoursToMinimum : float
        Number of hours to decrease before minimum credit.
        The default is 7*24.
    fullCreditSubs : integer
        The number of submissions that receive full credit.
        if zero or negative, then credit does not decrease with submissions.
    minCreditSubs : integer
        The number of submissions at which the decrease reaches the minimum
        credit for a correct answer.
        Must be greater than fullCreditSubs if fullCreditSubs is positive.
        Ignored if fullCreditSubs is zero or negative.
    minCreditFraction : float
        Fraction of the points received for correct answer with minCreditSubs
        submissions. If the credit decreases with time and submissions, then
        sqrt(minCreditFraction) is applied to time decrease and also to
        submission increase. The result is that after time runs out and
        submissions reaches minCreditSubs then the credit is reduced by this
        amount.
    assignmentDue: (a date string in format "%Y-%m-%d %H:%M:%S")
        due date/time for WebAssign assignment. This is the time that
        minute differences are calculated from.
    nudge: (in hours) the time to add to the decStart to allow for differences in
        time zones or computer clocks.

    Returns
    -------
    perl string to calculate the points, decreasing with time and with submissions

    '''
    
    '''
    There are three times needed:
        1. the time the assignment is due 
           ( assignmentDue --> epochDue --> zero minutes before due
        2. the time the credit starts decreasing 
           ( decStart) --> decMinutes = ( epochDue - epochStart)/60 ## minutes before due
        3. the time the credit has reached its minimum.
            ( decStart + hoursToMinimum) --> 
              endMinutes = ( epochDue - epochStart)/60 - hoursToMinimum*60 ## minutes before due
    
   '''
    

    outPerl = ''   ## string built up for output
   
    ### Determine factors for decrease with submissions and with time
    timeMult = 0.0
    subsMult = 0.0
    if decStart=='' or decStart==None:
        if fullCreditSubs<=0.0:
            print( 'ERROR neither decreases with time nor with submissions.')
            return None
        else:
            subsMult = minCreditFraction
    else:
        if fullCreditSubs<=0.0:
            timeMult = minCreditFraction
        else:
            timeMult = round( sqrt( minCreditFraction), 3)
            subsMult = round( minCreditFraction/timeMult, 3)
    # print( 'DBug: timeMult {}, subsMult {}'.format(
    #          timeMult, subsMult))
           
    if not ( decStart=='' or decStart== None):  ## if decrease with time.
    
    ## generate time decreasing part
        epochDue = localTimeToGMTepoch( assignmentDue)           ## in seconds
        epochStart = localTimeToGMTepoch( decStart) + nudge*3600 ## in seconds
        epochMinimum = epochStart + hoursToMinimum*3600          ## in seconds
        
        # print( 'DBug: epochDue {}, epochStart {}, epochMinimum {}'.format(
        #         epochDue, epochStart, epochMinimum))
        # print( 'DBug: (in s) Due-Start {}, Due-Minimim {}, Minimum-Start {}'.format(
        #         epochDue-epochStart, epochDue-epochMinimum, epochMinimum-epochStart))
        # print( 'DBug: (in min) Due-Start {}, Due-Minimim {}, Minimum-Start {}'.format(
        #         (epochDue-epochStart)/60, (epochDue-epochMinimum)/60,
        #         (epochMinimum-epochStart)/60))
        # print( 'DBug: (in hr) Due-Start {}, Due-Minimim {}, Minimum-Start {}'.format(
        #         (epochDue-epochStart)/3600, (epochDue-epochMinimum)/3600,
        #         (epochMinimum-epochStart)/3600))
        # print( 'DBug: (in day) Due-Start {}, Due-Minimim {}, Minimum-Start {}'.format(
        #         (epochDue-epochStart)/day, (epochDue-epochMinimum)/day,
        #         (epochMinimum-epochStart)/day))
        
        decMinutes = round( ( epochDue - epochStart)/60) ## minutes before due, when decrease starts
        endMinutes = round( ( epochDue - epochMinimum)/60) ## minumtes before due, when decrease ends
        
        # print( 'DBug decMinutes: {}, endMinutes: {}'.format(
        #        decMinutes, endMinutes))
        
        outPerl = "&cs9(-&beforeDue('minutes'),{},{},{})".format(
            -decMinutes, -endMinutes, timeMult)
        
        print( '\nAssignment is Due         {}, nudge is {} h.'.format(
              minBeforeToLocal( 00, assignmentDue), nudge))
        print( 'Decreasing start specified: {} Decrease time: {} h or {:0.2f} d'.format(
               decStart, hoursToMinimum, hoursToMinimum/24))
        print( 'Decreasing credit starts ',
              minBeforeToLocal( decMinutes, assignmentDue))
        print( 'Decreasing credit ends {}, to {}% credit, {}% dec. with time, {}% with subs.'.format(
              minBeforeToLocal( endMinutes, assignmentDue),
              minCreditFraction*100, timeMult*100, subsMult*100))
        
        ## add '*' if there is a submissions decrease part or finish line.
        if fullCreditSubs>0:
            outPerl = outPerl + '*\n'
        else:
            outPerl = outPerl + '*$POINTS</eqn>'
            
        # print( 'DBug: outPerl:\n{}'.format( outPerl))
    if not fullCreditSubs<=0:
        outPerl = outPerl + '&cs9($RESPONSE_NUM,{},{},{})*$POINTS</eqn>'.format(
                fullCreditSubs, minCreditSubs, subsMult)
    print( '\n++++++++++++++ Copy and Paste conditional code starting on next line.' +\
            '+++++++++++')
    print( '''<eqn>sub cs9{my($tn,$t1,$t2,$mn)=@_;
my $f=((1-$mn)*$tn+$mn*$t1-$t2)/($t1-$t2);
if($tn<=$t1){$f=1}
if($tn>=$t2){$f=$mn}
return $f}''')
    
    print( outPerl)
    # return outPerl


def BBPdecCredit( decStart, hoursToMinimum=7*24):
    '''
    Best By Prep decreasing credit
    Returns perl string that reduces credit after decStart.

    Convenience function that is equivalent to
      decCredit( decStart, hoursToMinimum, fullCreditSubs=5, minCreditSubs=10,
              minCreditFraction=0.2, assignmentDue=duedates[ 'PHYS250.001'], nudge)
        

    Parameters
    ----------
    decStart : (a date string in format "%Y-%m-%d %H:%M:%S")
        A date string, example:
            2020-08-31 23:59:00
    hoursToMin : TYPE
        Number of hours to decrease before minimum credit.

    Returns
    -------
    perl string to calculate the points, decreasing with time and with submissions

    '''
    decCredit( decStart, hoursToMinimum, fullCreditSubs=5, minCreditSubs=10,
              minCreditFraction=0.2, assignmentDue=duedates[ 'PHYS250.001'])

def InCdecCredit( decStart, hoursToMinimum=24):
    '''
    in class exercise
    Returns perl string that reduces credit after decStart.
    
    Convenience function that is equivalent to
      decCredit( decStart, hoursToMinimum, fullCreditSubs=0, minCreditSubs=10,
              minCreditFraction=0.2, assignmentDue=duedates[ 'PHYS250.001'], nudge)


    Parameters
    ----------
    decStart : (a date string in format "%Y-%m-%d %H:%M:%S")
        A date string, example:
            2020-08-31 23:59:00
    hoursToMin : TYPE
        Number of hours to decrease before minimum credit.

    Returns
    -------
    perl string to calculate the points, decreasing with time only.

    '''

    decCredit( decStart, hoursToMinimum, fullCreditSubs=0, minCreditSubs=10,
              minCreditFraction=0.4, assignmentDue=duedates[ 'PHYS250.001'])


def HWdecCredit( ):
    '''
    Home Work assignment, numerical answer

    Returns perl string that reduces credit after 5 submissions.
    
    Convenience function that is equivalent to
      decCredit( decStart, hoursToMinimum, fullCreditSubs=0, minCreditSubs=10,
              minCreditFraction=0.2, assignmentDue=duedates[ 'PHYS250.001'], nudge)
    

    Returns
    -------
    perl string to calculate the points, decreasing with time only.
    '''

    decCredit( '')

def is_number(s):
    '''
    

    Parameters
    ----------
    s : string
        string to check if it can be converted to a number.

    Returns
    -------
    bool
        Returns True if string can be converted to a number.
    Ref: https://stackoverflow.com/questions/354038/how-do-i-check-if-a-string-is-a-number-float
    '''
    try:
        float(s)
        return True
    except ValueError:
        return False

def is_datetime(s):
    '''
    Test if string is valid date-time string.
    Parameters
    ----------
    s : string
    Returns
    -------
    Boolean
    '''
    try:
        localTimeToGMTepoch( s) 
        return True
    except :
        return False


def getCCode():
    '''
    Prompts for arguments to decCredit() and then calls it
    to get Perl code for the conditional credit field in WebAssign.

    Returns
    -------
    outputs perl code.

    '''
    due = duedates[ "PHYS250.001"]
    fulldue = ""
    fullsubs = 5
    minsubs = 10
    tfrac = 0.4
    hours = 12
    
    print( '\nNote: Default values are in square brackets.\n Date format is {}'.format(
                due))
    prompt = 'Enter BestBy date, or <return> if credit does not decrease with time:'
    response = input( prompt.strip())
    
    if response == "":
        pass
    elif is_datetime( response):
        fulldue = response
    else:
        print( '{} is not in proper Date Time format.'.format( response))
        return

    
    if not fulldue == "":
        ## ck if due date is default
        prompt = 'Enter WebAssign due date [{}]:'
        response = input( prompt.format( duedates[ "PHYS250.001"])).strip()
    
        if response == "":
            pass
        elif is_datetime( response):
            # due = localTimeToGMTepoch( response) 
            due = response
        else:
            print( '{} is not in proper Date Time format.'.format( response))
            return
        print( 'Due: {}'.format( due))
    
        
    

    prompt = 'Enter the number of submissions that receive full credit,\n 0 for all at full credit [{}]:'      
    response = input( prompt.format( fullsubs)).strip()
    
    if response == "":
        pass
    elif response == 'zero' or response == '0':
        fullsubs = 0
        minsubs = 0
        tfrac = 0.4
    elif not is_number( response):
        print( '{} is not a number.'.format( response))
        return
    elif float( response) <= 0.0:
        fullsubs = 0
        minsubs = 0
    elif int( response) > 0:
        fullsubs = int( response)
    else:
        print( 'Did not recognize "{}" for response.'.format( response))
        return
    
    if fullsubs > 0:
        prompt = "Enter number of submissions where credit reaches minimum [{}]:"
        response = input( prompt.format( minsubs)).strip()
        
        if response == "":
            pass
        elif not is_number( response):
            print( '{} is not a number.'.format( response))
            return
        elif int( response) >= fullsubs:
            minsubs = int( response)
        else:
            print( 'Did not recognize "{}" for response.'.format( response))
            return
            
    if fulldue != "":  ## credit decreases with time, need to know how fast.
        prompt = 'Enter number days for credit to decrease to minumum [{:.2f}]:'    
        response = input( prompt.format( hours/24)).strip()
        # print('DBug response: {}'.format( response))
        if response == "":
            pass
        elif not is_number( response):
            print( '{} is not a number.'.format( response))
            return
        else:
            hours = float( response)*24
        prompt = 'Enter fractional credit after time and submission limits [{}]:'
        response = input( prompt.format( tfrac)).strip()
        if response == "":
            pass
        elif not is_number( response):
            print( '{} is not a number.'.format( response))
            return
        else:
            tfrac = float( response)
            if tfrac > 1.0 or tfrac < 0.0:
                print( '{} is not a valid fractional decrease.'.format( tfrac))
                return
 

    # print( 'DBug: fulldue "{}", due {}, fullsubs {}, minsubs {}, tfrac {}, hours {}'.format(
    #     fulldue, due, fullsubs, minsubs, tfrac, hours))
    
    print( '_____________  Perl Code Below ++++++++++++++')
    
    decCredit( fulldue, hours, fullsubs, minsubs, tfrac, due)
    return

if __name__=="__main__":
    print( '\nTesting BestByMinBefore.py\n')
    
    decStart = '2020-09-01 23:59'
    print( 'decStart:', decStart)
    
    print( '\ndecCredit:')
    decCredit( decStart)
    
    print( '\nHW:')
    HWdecCredit()
    
    print( '\nInCdecCredit:')
    InCdecCredit( decStart)
    
    print( '\nBBPdecCredit:')
    BBPdecCredit( decStart)
    
    
    
    # test1 = decCredit( '2020-08-21 23:58:00')
    # assert test1=="&cs9(-&beforeDue('minutes'),-169867,-159787,0.447)*\n&cs9($RESPONSE_NUM,5,10,0.447)*$POINTS</eqn>"
    
    decStart = '2020-09-01 23:59'
    print( 'decStart:', decStart)
    
    print( '\ndecCredit:')
    decCredit( decStart)
    
    print( '\nHW:')
    HWdecCredit()
    
    print( '\nInCdecCredit:')
    InCdecCredit( decStart)
    
    print( '\nBBPdecCredit:')
    BBPdecCredit( decStart)
    
    
    
    # test1 = decCredit( '2020-08-21 23:58:00')
    # assert test1=="&cs9(-&beforeDue('minutes'),-169867,-159787,0.447)*\n&cs9($RESPONSE_NUM,5,10,0.447)*$POINTS</eqn>"
    
    


import setuptools
import codecs
import os.path

##   ---- Time-stamp: <2020-11-12T09:43:25.299646-05:00 hedfp> -----

## code to get version number from __init__.py

def read(rel_path):
    '''Reads text file.'''
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    '''Reads version from specified file. The file path is relative
    to this file. It looks for line that starts with "__version__"'''
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")
## end version number extraction functions


def timeStampStr():
    '''Returns a string with the current time in ISO format.'''
    import datetime
    import time
    
    dt = datetime.datetime.now()
    epoch = dt.timestamp() # Get POSIX timestamp of the specified datetime.
    st_time = time.localtime(epoch) #  Get struct_time for the timestamp. This will be created using the system's locale and it's time zone information.
    tz = datetime.timezone(datetime.timedelta(seconds = st_time.tm_gmtoff)) # Create a timezone object with the computed offset in the struct_time.
    return dt.astimezone(tz).isoformat()


def writeTS( rel_path):
    '''Writes the current time and date to specified file. The file path is
    relative to this file.'''
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'w') as tsf:
        tsf.write( timeStampStr())
        tsf.close()

def getTS( rel_path):
    '''Reads first line of specified file, which is expected to be a time-date
    string.'''
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as tsf:
        lines = tsf.read()
    linelist = lines.splitlines()
    return linelist[0]

## Run tests here??

## Save timestamp for update
writeTS(( 'classroom_gizmos/timestamp.txt'))

## Get description from README.md.
with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="classroom_gizmos",
    ####version="0.0b1.dev5", # X.YaN.devM format
    version=get_version("classroom_gizmos/__init__.py"),
    author="Carl Schmiedekamp",
    author_email="cw2@psu.edu",
    description="Several small functions for classroom instruction.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Danseur1/Classroom_gizmos",
    package_data={ ## See: https://packaging.python.org/guides/distributing-packages-using-setuptools/#package-data
        # all *.txt files
        "": ["*.txt"]} ,
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
	"License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
	"Development Status :: 3 - Alpha"
    ],
    python_requires='>=3.6',
)
